/************Server filialeRead/Write.c****************************** 
Fri Jan 15 21:35:43 1999

Azioni del programma:
1)  Controllo sull'esattezza dei parametri in ingresso
2) Registrazione sul server del servizio
3) Attesa di connessioni da parte dei client
Utilizzo la libreria NDBM x l'accesso ai dati */

#define BUFSIZE 100
#define DNSPORT	 32001
#define DNSPORT2 32000

#include <stdio.h>
#include <ndbm.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
/*Prototipo*/
void uscita();


struct hostent *hostdns,*hfiliale; /*Puntatore all'host remoto*/
struct sockaddr_in myaddr_in; /* Socket address locale*/
struct sockaddr_in peeraddr_in,dnsaddr_in; 

DBM * dbf;
datum key,content;
datum datoric;
int ret,s,lunghezza,WR;
char * testo;
char * chiave;
char *nomefiliale;
char buf[BUFSIZE];

main(argc,argv)
int argc; char **argv;
{
int addrlen,len,i;
WR=0;/*Variabile che indica se il server sara' lettore o scrittore*/
/* Check sul numero di parametri in ingresso*/
if (argc!=5)
    {
    fprintf(stderr,"\nUso:%s  <server di nomi> <nome filiale> <porta> <servizio>\n",argv[0]);
    fprintf(stderr,"<server di nomi>: Nome del Name Server in /etc/hosts\n");
    fprintf(stderr,"<nome filiale>: Nome logico sul name server\n");
     fprintf(stderr,"<porta>: Porta da utilizzare per fornire il servizio\n");
     fprintf(stderr,"<servizio>: Tipo di servizio da offrire sulla base dati: [read/write]\n\n");
    exit(1);
    }

/*Check sul Name Server*/
if ((hostdns=gethostbyname(argv[1]))==NULL)
    {
    fprintf(stderr,"%s: %s non trovato in /etc/hosts\n",argv[0],argv[1]);
    fprintf(stderr,"Uso:%s  <server di nomi> <nome filiale> <porta> <servizio>\n\n",argv[0]);
    exit(1);
    }  

/*Check sul servizio da offrire sulla base dati*/
if(strcmp(argv[4],"read")==0) WR=0;
    else if(strcmp(argv[4],"write")==0) WR=1;
		else
		    {
	             fprintf(stderr,"%s: Parametro non valido: %s, usare [read/write]\n",argv[0],argv[4]);
		     fprintf(stderr,"Uso:%s  <server di nomi> <nome filiale> <porta> <servizio>\n\n",argv[0]);
		    exit(1);
		    }  
	    
/*Apro il database in lettura a seconda di WR, con nome argv[2]+".db"*/
strcpy(buf,argv[2]);
strcat(buf,".db");
if(!WR) dbf=dbm_open(buf,O_RDONLY,0666);/* Nel caso si scelga l'azione di lettura del database*/

if ((dbf==NULL)&&(WR==0))
    {
    perror(argv[0]);
    fprintf(stderr,"%s: Errore nell'apertura del file database.\n",argv[0]);
    free(nomefiliale);
    close(s);
    exit(1);
    }
    
/*Copio il nome della filiale in un'altra stringa*/
nomefiliale=(char *)malloc(sizeof(argv[2])+1);
strcpy(nomefiliale,argv[2]);

/* Azzera le strutture degli indirizzi*/
memset((char *)&myaddr_in,0,sizeof(struct sockaddr_in));
memset((char *)&dnsaddr_in,0,sizeof(struct sockaddr_in));
memset((char *)&peeraddr_in,0,sizeof(struct sockaddr_in));



/* Assegnazioni per la bind della socket*/
myaddr_in.sin_family=AF_INET;
myaddr_in.sin_addr.s_addr=INADDR_ANY;/*OSS: Il server ascolta su qualunque interfaccia di rete, invece che su una in particolare: utile nel caso di piu' schede di rete*/
myaddr_in.sin_port=htons(atoi(argv[3]));/*Assegniamo la porta di richiesta*/

/* Riempiamo la struttura sockaddr per registrarci sul NS*/
dnsaddr_in.sin_family=AF_INET;
dnsaddr_in.sin_addr.s_addr=((struct in_addr *)(hostdns->h_addr))->s_addr;
dnsaddr_in.sin_port=htons(DNSPORT); /*Porta di default del NS*/

/* Creazione della socket*/
s=socket(AF_INET,SOCK_DGRAM,0);
if (s==-1)
    {
    perror("Errore apertura socket");
    fprintf(stderr,"Non posso creare la socket\n");
    free(nomefiliale);
    exit(1);
    }

/*Faccio la bind*/
if (bind(s,&myaddr_in,sizeof(struct sockaddr_in))==-1)
    {
    perror("Errore nella bind");
    fprintf(stderr,"Non posso attaccarmi alla porta%s\n",argv[3]);
    free(nomefiliale);
    exit(1);
    }


/* Invio al Name Server*/
 /*il comando 1 che corrisponde alla registrazione di servizio.
La forma della richiesta e':
'1'+nomefiliale se lettore
'1'+nomefiliale+'W' se scrittore*/

buf[0]='1';
strcpy(&buf[1],nomefiliale);
if(WR) strcat(buf,"W");

if (sendto(s,buf,(strlen(buf)+1),0,&dnsaddr_in,sizeof(dnsaddr_in))==-1)
	{
	perror("Query terminata per errore");
	fprintf(stderr,"sulla richiesta al NS di registrazione: %s\n",nomefiliale);
	free(nomefiliale);
	exit(1);
	}

printf("Debug: messaggio al NS  inviato\n");

/*Ricezione del messaggio dal NS contenente la conferma*/
if (recvfrom(s,buf,BUFSIZE,0,&dnsaddr_in,&lunghezza)==-1)
    {
    perror("Errore in ricezione dal NS");
    fprintf(stderr,"sulla richiesta al NS del server: %s\n",nomefiliale);
    free(nomefiliale);
    exit(1);
    }

/*Esaminiamo il buffer ricevuto*/
if (strcmp("Ok1",buf)!=0)
    {
    fprintf(stderr,"Nome della filiale non registrato sul NS\n");
    free(nomefiliale);
    close(s);
    exit(1);
    }
    else printf("DEBUG: Registrazione effettuata correttamente\n");

/*Istallo l'handler per gestire l'uscita*/
signal(SIGINT,uscita);/*Gestisce il CTRL-C */

/*Mi metto in attesa di richieste dai client*/
for(;;) 
	{
	addrlen=sizeof(struct sockaddr_in);
	if((len=recvfrom(s,buf,sizeof(buf),0,&peeraddr_in,&addrlen))==-1)
	    { 
	    printf("Errore nella RECVFROM\n");
	    if (!WR) dbm_close(dbf);
	    close(s);
	    exit(1);
	    }
	    else	if (WR) ret=serverW();/* Per ogni chiamata viene invocata la funzione server x gestire il servizio*/								    
			    else ret=serverR();/*lettore*/
	/*DEV: Da gestire!!*/
	} /* Fine del for*/
}/*Fine main()*/

/* Procedura serverW()*/
/*Il pacchetto di ricezione che arriva dai processi terminali ha la seguente struttura:
1) il primo carattere contiene il servizio richiesto e puo'  assumere i seguenti valori:
	    1 Inserimento dato
	    2 Cancellazione dato
	    3 Sostituzione
2) Segue una stringa di dati terminata '\0' che contiene la chiave del database
3) Segue una stringa con gli eventuali dati da memorizzare
*/
	     
	
int serverW()
{
char buf2[20];
int lunghezza2,len;

buf2[0]='5';
strcpy(&buf2[1],nomefiliale);

/*Poiche' la libreria ndbm non permette l'accesso in scrittura al file se ha piu' istanze,
invio al Name Server la richiesta di apertura del database  in modo esclusivo*/
if (sendto(s,buf2,(strlen(buf2)+1),0,&dnsaddr_in,sizeof(dnsaddr_in))==-1)
	{
	perror("Invio terminata per errore");
	fprintf(stderr,"sulla richiesta al DNS di registrazione esclusiva: %s\n",nomefiliale);
	free(nomefiliale);
	exit(1);
	}

printf("Debug: messaggio al NS  inviato\n");

/*Ricezione del messaggio dal NS contenente la conferma x l'apertura*/
if (recvfrom(s,buf2,20,0,&dnsaddr_in,&lunghezza)==-1)
    {
    perror("Errore in ricezione dal NS");
    fprintf(stderr,"sulla richiesta al NS del server: %s\n",nomefiliale);
    free(nomefiliale);
    exit(1);
    }

/*Esaminiamo il buffer ricevuto*/
if (strcmp("Ok5",buf2)!=0)
    {
    fprintf(stderr,"Nome della filiale non registrato sul DNS\n");
    free(nomefiliale);
    close(s);
    exit(1);
    }
    else printf("DEBUG: Registrazione effettuata correttamente\n");

/*So che a questo punto il database puo' essere aperto senza problemi*/
strcat(nomefiliale,".db");
dbf=dbm_open(nomefiliale,O_WRONLY|O_CREAT,0666);
if (dbf==NULL)
    {
    fprintf(stderr,"Errore nell'apertura del file database in scrittura.\n");
    free(nomefiliale);
    close(s);
    exit(1);
    }
len=strlen(nomefiliale);
nomefiliale[len-3]='\0';

/*il buffer e' nella forma: '1'+Cognome+Nome+'\0'+campo dati+'\0'*/
/*Riempio la strutture key */
lunghezza=strlen(&buf[1])+1;/*Lunghezza Cognome+Nome*/
key.dptr=(char *)malloc(lunghezza);
strcpy(key.dptr,&buf[1]);
key.dsize=lunghezza;
/*...e content nel caso fosse presente*/
if(buf[0]!='2')
    {
    lunghezza2=strlen(&buf[1+lunghezza])+1;
    content.dptr=(char *)malloc(lunghezza2);
    strcpy(content.dptr,&buf[lunghezza+1]);
    content.dsize=lunghezza2;
    }

/*Decodifica e esecuzione dell'azione richiesta*/
switch(atoi(&buf[0]))
    {
    case 1:/* Inserisco un dato */
		ret=dbm_store(dbf,key,content,DBM_INSERT);
		    switch(ret)
			    {
			    case -1: printf("Errore nell'inserimento nel database!!!\n");
					    free(content.dptr);
					    free(key.dptr);
					    break;
			    case 1:printf("Errore nell'inserimento nel database!!! Chiave gia' presente\n");
					strcpy(buf,"Er1");
					if((sendto(s,buf,(strlen(buf)+1),0,&peeraddr_in,sizeof(peeraddr_in)))==-1)
					    {
					    perror("Errore nella conferma");
					    fprintf(stderr,"Errore nella send di conferma in serverW()");
					    break; 
					    } 
					free(content.dptr);
					free(key.dptr);
					break;
			    default:printf("Inserito correttamente nel database\n");
					/*Notifico l'avvenuto inserimento*/
					free(content.dptr);
					free(key.dptr);
					strcpy(buf,"Ok1");
					if((sendto(s,buf,(strlen(buf)+1),0,&peeraddr_in,sizeof(peeraddr_in)))==-1)
					    {
					    perror("Errore nella conferma");
					    fprintf(stderr,"Errore nella send di conferma in serverW()");
					    }
					break; 
			    }
		 break;/*Tutto Ok*/
		 
    case 2:/*Cancello il dato*/
		ret=dbm_delete(dbf,key);
		    switch(ret)
			    {
			    case -1: printf("Elemento da cancellare non presente nel database!!!\n");
					    free(key.dptr);
					    strcpy(buf,"Er2");
					    if((sendto(s,buf,(strlen(buf)+1),0,&peeraddr_in,sizeof(peeraddr_in)))==-1)
						{
						perror("Errore nella conferma");
						fprintf(stderr,"Errore nella send di conferma in serverW()");
						} 
					    break;
			    default:printf("Cancellato correttamente dal database\n");
					/*Notifico l'avvenuta cancellazione*/
					free(key.dptr);
					strcpy(buf,"Ok2");
					if((sendto(s,buf,(strlen(buf)+1),0,&peeraddr_in,sizeof(peeraddr_in)))==-1)
					    {
					    perror("Errore nella conferma");
					    fprintf(stderr,"Errore nella send di conferma in serverW()");
					    } 
			    }
		break;/*Tutto Ok*/
			    
    case 3:/*Sostituisco un dato */
		if((ret=dbm_store(dbf,key,content,DBM_REPLACE))==0)
			{
			printf("Dato modificato correttamente \n");
			/*Notifico l'avvenuto inserimento*/
			free(content.dptr);
			free(key.dptr);
			strcpy(buf,"Ok3");
			if((sendto(s,buf,(strlen(buf)+1),0,&peeraddr_in,sizeof(peeraddr_in)))==-1)
				    {
				    perror("Errore nella conferma");
				    fprintf(stderr,"Errore nella send di conferma in serverW()");
				    } 
			
			}
			else
			    {
			    printf("Errore nella modifica del database!!!\n");
			    strcpy(buf,"Er3");
			    if((sendto(s,buf,(strlen(buf)+1),0,&peeraddr_in,sizeof(peeraddr_in)))==-1)
				    {
				    perror("Errore nella conferma");
				    fprintf(stderr,"Errore nella send di conferma in serverW()");
				    break; 
				    } 
			    break;
			    }
		break;/*Tutto OK*/	
		
    default: break;/*Ignora la richiesta*/
    }/*Fine switch*/

/*Chiudo il database*/dbm_close(dbf);

/*Notifico al Name Server che l'azione di scrittura e' terminata*/
strcpy(buf,"Start");
dnsaddr_in.sin_port=htons(DNSPORT2);
if (sendto(s,buf,(strlen(buf)+1),0,&dnsaddr_in,sizeof(dnsaddr_in))==-1)
	{
	perror("Invio terminata per errore");
	fprintf(stderr,"sulla richiesta al NS di registrazione esclusiva: %s\n",nomefiliale);
	free(nomefiliale);
	exit(1);
	}
	
/*Riperistino il valore di porta corretta x il dns*/
dnsaddr_in.sin_port=htons(DNSPORT); /*Porta di default del NS*/
}




/*Funzione server lettore
I comandi che accetta sono:
5:Viene chiuso il database aperto in lettura, per consentire un accesso 
  esclusivo in scrittura
1:restituisce il valore del conto corrente*/


int serverR()
{
int lunghezza,lunghezza2;
char buf2[20];

lunghezza=sizeof(peeraddr_in);
/*Caso di accesso al file in scrittura da parte di un server scrittore*/
/*E' necessario che tutti i server chiudano il file aperto*/
if (buf[0]=='5')
    {
    dbm_close(dbf); /*Chiudo il database*/
    strcpy(buf2,"Ok5");
    if (sendto(s,buf2,(strlen(buf2)+1),0,&peeraddr_in,sizeof(peeraddr_in))==-1)
	{
	perror("Errore in server() 5");
	fprintf(stderr,"sulla conferma al NS di chiusura del file: %s\n",nomefiliale);
	free(nomefiliale);
	exit(1);
	}
    if (recvfrom(s,buf2,20,0,&peeraddr_in,&lunghezza)==-1)
	{
	perror("Errore in ricezione dal NS");
	fprintf(stderr,"sulla richiesta al NS del server: %s\n",nomefiliale);
	free(nomefiliale);
	exit(1);
	}
    if(strcmp(buf2,"5Restart")==0) 
	{
	strcpy(buf,nomefiliale);
	strcat(buf,".db");
	/*Ora posso riaprire il file in lettura*/
	dbf=dbm_open(buf,O_RDONLY,0666);
	
	if (dbf==NULL)
	    {
	    fprintf(stderr,"Errore nella riapertura del file database.\n");
            free(nomefiliale);
	    close(s);
	    exit(1);
	    }
	return(0);
	}
	else
	    {
	    printf("Errore nella ricezione del comando Restart\n");
	    exit(1);
	    }
	    
    }/*Fine if*/
	
	
/*Riempio la struttura key */
lunghezza=strlen(buf);/*Lunghezza Cognome+Nome+terminatore*/
key.dptr=(char *)malloc(lunghezza);
strcpy(key.dptr,&buf[1]);
key.dsize=lunghezza;

/*Decodifica e esecuzione dell'azione richiesta*/
switch(atoi(&buf[0]))
    {
    case 1:/* Visualizza un dato */
		datoric=dbm_fetch(dbf,key);/*Alloca la memoria di datoric.dptr*/
		if(datoric.dptr==NULL)
			{
			 /*printf("Errore nella ricerca nel database!!!\n");*/
			 strcpy(buf,"Er1");
					if((sendto(s,buf,(strlen(buf)+1),0,&peeraddr_in,sizeof(peeraddr_in)))==-1)
					{
					perror("Errore nella conferma");
					fprintf(stderr,"Errore nella send di conferma in serverR()");
					return (-2); 
					} 
			free(key.dptr);
			return(0);
			}
		else
		    {
		    printf("Elemento trovato\n");
		    strcpy(buf,datoric.dptr);
		    if((sendto(s,buf,(strlen(buf)+1),0,&peeraddr_in,sizeof(peeraddr_in)))==-1)
			{
			perror("Errore nella conferma");
			fprintf(stderr,"Errore nella send di conferma in serverR()");
			free(key.dptr);
			/*free(datoric.dptr);*/
			return (-2); 
			} 
	    	    /*free(datoric.dptr);*/
		    free(key.dptr);
		    return(0);
		    }
		
    default:return(-3);/*Ignoro la richiesta*/
    }
}
	
/*Handler  di cancellazione dal Name Server*/
void uscita()
{
/*Chiudo il database*/
if (!WR) dbm_close(dbf);

/*Imposto  il servizio da che deve essere richiesto al NS*/
buf[0]='2';
strcpy(&buf[1],nomefiliale);
if(WR) strcat(buf,"W");

/*send al NS*/
if (sendto(s,buf,(strlen(buf)+1),0,&dnsaddr_in,sizeof(dnsaddr_in))==-1)
	{
	perror("Cancellazione terminata per errore");
	fprintf(stderr,"sulla richiesta al NS di registrazione: %s\n",nomefiliale);
	exit(1);
	}

printf("Debug: messaggio  inviato\n");

/*Ricezione del messaggio dal NS di conferma chiusura*/
if (recvfrom(s,buf,100,0,&dnsaddr_in,&lunghezza)==-1)
    {
    perror("Errore in ricezione dal NS");
    fprintf(stderr,"sulla richiesta al NS del server: %s\n",nomefiliale);
    exit(1);
    }

/*Esaminiamo il buffer ricevuto*/
if (strcmp("Ok2",buf)!=0)
    {
    printf("Nome della filiale non registrato sul NS\n");
    close(s);
    exit(1);
    }

/*Chiudo la socket*/
free(nomefiliale);
close(s);
exit(0);
}



